(***************************************************************************
* Bundles of Modules for Developments in Programming Language Metatheory   *
* Arthur Charguéraud, July 2007, Coq v8.1pl4                               *
***************************************************************************)

Set Implicit Arguments.
Require Import LibTactics List.


Tactic Notation "gen_eq" ident(X1) ":" constr(E1) ","
  ident(X2) ":" constr(E2) :=
  gen_eq X2: E2; gen_eq X1: E1.
Tactic Notation "gen_eq" ident(X1) ":" constr(E1) ","
  ident(X2) ":" constr(E2) "," ident(X3) ":" constr(E3) :=
  gen_eq X3: E3; gen_eq X2: E2; gen_eq X1: E1.

Tactic Notation "inductions" constr(E) "with" tactic(GEN) "," tactic(INTROS) :=
  GEN; induction E; INTROS.
Tactic Notation "inductions" constr(E) "with" "," tactic(INTROS) :=
  induction E; INTROS.

Tactic Notation "intro_subst" :=
  let H := fresh in intros H; 
  match type of H with
  | ?x = ?y => first [ subst x | subst y ] 
  | _ => fail 1 ">- you probably forgot to perform some introductions" 
  end.

Tactic Notation "inductions" constr(E) "with"
  tactic(GEN) "," tactic(INTROS) "end" :=
  GEN; induction E; INTROS.

Tactic Notation "inductions" constr(E) "with"
  "gen_eq" ident(X1) ":" constr(E1) "and"
  tactic(GEN) "," tactic(INTROS) "end" :=
  gen_eq X1: E1; GEN; induction E; INTROS; try intro_subst.

Tactic Notation "inductions" constr(E) "with"
  "gen_eq" ident(X1) ":" constr(E1) ","
           ident(X2) ":" constr(E2) "and"
  tactic(GEN) "," tactic(INTROS) "end" :=
  gen_eq X2: E2; gen_eq X1: E1; GEN; induction E; INTROS; 
   try (do 2 intro_subst).


Tactic Notation "inductions" constr(E) "with"
  tactic(GEN) "end" :=
  inductions E with (GEN), (idtac).

Tactic Notation "inductions" constr(E) "with"
  "gen_eq" ident(X1) ":" constr(E1) "and"
  tactic(GEN) "end" :=
  inductions E with gen_eq X1: E1 and (GEN), (idtac) end.

Tactic Notation "inductions" constr(E)  "with"
  "gen_eq" ident(X1) ":" constr(E1) ","
           ident(X2) ":" constr(E2) "and"
  tactic(GEN) "end" :=
  inductions E with gen_eq X1: E1, X2:E2 and (GEN), (idtac) end.


Tactic Notation "inductions" constr(E) :=
  inductions E with (idtac), (idtac).

Tactic Notation "inductions" constr(E) "with"
  "gen_eq" ident(X1) ":" constr(E1) :=
  inductions E with gen_eq X1: E1 and (idtac), (idtac) end.

Tactic Notation "inductions" constr(E) "with" 
  "gen_eq" ident(X1) ":" constr(E1) "," 
           ident(X2) ":" constr(E2) :=
  inductions E with gen_eq X1: E1, X2:E2 and (idtac), (idtac) end.

(*  in stlc_core
Lemma demo_inductions : forall G E F t T,
   (E & G) |= t ~: T -> 
   ok (E & F & G) ->
   (E & F & G) |= t ~: T.
Proof.
  introv Typ. dup 3.
  gen_eq H: (E & G). induction Typ; introv EQ; subst; skip.
  inductions Typ with gen_eq H:(E&G); skip.
  inductions Typ with gen_eq H:(E&G) and gen F, introv end; skip.
Qed.
*)


Tactic Notation "simple" :=
  simpl.
Tactic Notation "simple" "~" :=
  simpl; auto_tilde.
Tactic Notation "simple" "*" :=
  simpl; auto_star.


(*todo;move*)
Tactic Notation "tryfalse" "~" :=
  try solve [ false; auto_tilde ].
Tactic Notation "tryfalse" "*" :=
  try solve [ false; auto_star ].

(* fix 
Ltac get_term_disjunction_arity T :=
  match type of T with
*)


(** [do_rew] is used to perform the sequence: 
  rewrite the goal, execute a tactic, rewrite the goal back *)

Tactic Notation "do_rew" constr(E) tactic(T) :=
  rewrite E; T; try rewrite <- E.

Tactic Notation "do_rew" "<-" constr(E) tactic(T) :=
  rewrite <- E; T; try rewrite E.

Tactic Notation "do_rew" "*" constr(E) tactic(T) :=
  rewrite E; T; auto*; try rewrite* <- E.
Tactic Notation "do_rew" "*" "<-" constr(E) tactic(T) :=
 rewrite <- E; T; auto*; try rewrite* E.

(** [do_rew_2] is the same as [do_rew] but it does rewrite twice *)

Tactic Notation "do_rew_2" constr(E) tactic(T) :=
  do 2 rewrite E; T; try do 2 rewrite <- E.

Tactic Notation "do_rew_2" "<-" constr(E) tactic(T) :=
  do 2 rewrite <- E; T; try do 2 rewrite E.


(** Short-hand notations for projections from triples and quadruples. *)

Notation "'proj31' P" := (proj1 P) (at level 69).
Notation "'proj32' P" := (proj1 (proj2 P)) (at level 69).
Notation "'proj33' P" := (proj2 (proj2 P)) (at level 69).

Notation "'proj41' P" := (proj1 P) (at level 69).
Notation "'proj42' P" := (proj1 (proj2 P)) (at level 69).
Notation "'proj43' P" := (proj1 (proj2 (proj2 P))) (at level 69).
Notation "'proj44' P" := (proj2 (proj2 (proj2 P))) (at level 69).

(* todo: move *)


(** Results on List.map *)

Lemma list_map_id : forall (A : Set) l,
  List.map (fun t : A => t) l = l.
Proof.
  induction l; simpl; f_equal*.
Qed.

Lemma list_concat_right : forall (A : Set) (x : A) l1 l2,
  l1 ++ (x :: l2) = (l1 ++ (x :: nil)) ++ l2.
Proof.
  intros. induction l1; simpl; f_equal*.
Qed.

Lemma list_map_nth : forall (A : Set) (f : A -> A) d l n,
  f d = d -> 
  f (List.nth n l d) = List.nth n (List.map f l) d.
Proof.
  induction l; simpl; introv Fix; destruct* n.
Qed.

(** Results on List.forall *)

Inductive list_forall (A : Set) (P : A -> Prop) : list A -> Prop :=
  | list_forall_nil : list_forall P nil
  | list_forall_cons : forall L x,
       list_forall P L -> P x -> list_forall P (x::L).

Hint Constructors list_forall.

Lemma list_forall_concat : forall (A : Set) (P : A -> Prop) L1 L2,
  list_forall P L1 -> 
  list_forall P L2 ->
  list_forall P (L1 ++ L2).
Proof.
  induction L1; simpl; intros. auto. inversions* H.
Qed.

(** Results on List.for_n *)

Definition list_for_n (A : Set) (P : A -> Prop) (n : nat) (L : list A) :=
  n = length L /\ list_forall P L.

Lemma list_for_n_concat : forall (A : Set) (P : A -> Prop) n1 n2 L1 L2,
  list_for_n P n1 L1 -> 
  list_for_n P n2 L2 ->
  list_for_n P (n1+n2) (L1 ++ L2).
Proof.
  unfold list_for_n. intros. split. 
  rewrite* app_length. apply* list_forall_concat.
Qed.

Hint Extern 1 (?n = length ?xs) => 
 match goal with H: list_for_n _ ?n ?xs |- _ => 
  apply (proj1 H) end.

Hint Extern 1 (length ?xs = ?n) => 
 match goal with H: list_for_n _ ?n ?xs |- _ => 
  apply (sym_eq (proj1 H)) end.







